################################################################################
#                                  EMMA_ap.py                                 #
################################################################################
# Copyright (C) 2013 Matthew Dyer and Christopher Collins                      #
#   This program is free software: you can redistribute it and/or modify       #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   This program is distributed in the hope that it will be useful,            #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.      #
################################################################################
from ase.all import *
import numpy
import sys
import os
import time
import math

def order_atoms(atoms):
    """Function to order atoms object according to atomic number."""
    Z=atoms.get_atomic_numbers()
    sort_index=numpy.argsort(Z)
    sort_index=list(sort_index)
    new_atoms=atoms.copy()[sort_index]
    return new_atoms

def make_permutations(in_array):
    """Function which returns permutations of array."""
    from itertools import permutations
    perms=[]
    for array in permutations(in_array):
        perms.append(array)
    numpy.array(perms)
    return numpy.unique(perms)


def write_structures(all_perms,structures,write_format,labelling,shells=None):
    """Function which writes out structures."""
    for istruct in range(len(all_perms)):
        arrangement=all_perms[istruct]
        structure=structures[istruct]
        #Create a label for this arrangement
        if labelling.lower()=='numeric':
            label = '%d'%int(numpy.argwhere(numpy.all(numpy.all(all_perms==arrangement,axis=-1),axis=-1)))
        elif labelling.lower()=='padded':
        #Label numerically
            #Work out maximum necessary length of integer string
            int_length=math.ceil(math.log10(len(all_perms)))
            format_specifier='%%0%dd'%int_length
            label = format_specifier%int(numpy.argwhere(numpy.all(numpy.all(all_perms==arrangement,axis=-1),axis=-1)))
        elif labelling.lower()=='vasp neb':
        #Label numerically so that a VASP NEB type calculation can be performed
        #In this case there should not be a directory called 00 and maximum number 
        #of structures is 98
            #Work out if we have too many structures
            if len(all_perms)>98:
                print "Cannot use VASP NEB style labelling since there more than 98 structures"
            label = '%02d'%(int(numpy.argwhere(numpy.all(numpy.all(all_perms==arrangement,axis=-1),axis=-1)))+1)
        else:
        #Label according to modular structure
            label = ''
            for iblock in range(nblocks):
                label = label + 'A%d'%arrangement[iblock,0]
                label = label + 'B%d'%arrangement[iblock,1]
        #Write out geometry
        if write_format.lower()=='vasp':
            #Create new directory for structure if it doesn't exist
            if not os.path.exists(label):
                os.mkdir(label)
            write(label+"/POSCAR",structure,sort=False)
        elif write_format.lower()=='gulp':
            #Create new directory for structure if it doesn't exist
            if not os.path.exists(label):
                os.mkdir(label)
            filename=label+"/"+label+".res"
            write_res(filename,structure,name=label,shells=shells)

def make_structures(AMods,BMods,nA,nB,write_format='',labelling='numeric',
                    shells=None,rattle_dist=0.):
    """Function to make all possible structures from modules.
 
       nblocks is the number of each type of block in the final structure. 
       The final structures are then 2*nblocks high.
       AMods are the separate building blocks which are A-site terminated
       BMods are the separate building blocks which are B-site terminated
       The number of each block should be stated in the nA and nB arrays  
       which must therefore be the same length as AMods and BMods.
       Note that the total number of A and B-site terminated blocks (nblocks)
       must be the same
       rattle_dist is the standard deviation for a random rattling of the 
                   atomic positions to avoid symmetric starting structures
    """
    #Test lengths of arrays
    nA=numpy.array(nA)
    nB=numpy.array(nB)
    if nA.sum()!=nB.sum():
        print "Error: Total number of A and B terminated blocks must be the same."
        sys.exit()
    else:
        nblocks=nA.sum()
    #Get possible permutations of A and B modules
    print "Building array of all possible structures" 
    starttime=time.time()
    Aarray=[]
    for imod in range(len(AMods)):
        for irepeat in range(nA[imod]):
            Aarray.append(imod)
    Aarray=numpy.array(Aarray)
    Barray=[]
    for imod in range(len(BMods)):
        for irepeat in range(nB[imod]):
            Barray.append(imod)
    Barray=numpy.array(Barray)
    Aperms=make_permutations(Aarray)
    Bperms=make_permutations(Barray)
    all_perms=numpy.zeros([len(Aperms)*len(Bperms),nblocks,2],numpy.integer)
    all_perms[:,:,0]=numpy.repeat(Aperms,len(Bperms),axis=0)
    all_perms[:,:,1]=numpy.tile(Bperms,(len(Aperms),1))
    endtime=time.time()
    print "Time taken %10.5f s" % (endtime-starttime)
    print "Removing periodic repeats from %d structures" % len(all_perms)
    #Identify and remove periodic repeats. 
    #Loop over unique arrangements
    #Slow way
    #starttime=time.time()
    #iunique=0
    #length=len(all_perms)
    #while iunique<length:
    #    unique_perm=all_perms[iunique]
    #   #Find periodic translations of this arrangement
    #   #and remove them from the original array
    #    for itranslate in range(1,nblocks):
    #       #Set up indices in reverse order from current unique arrangement to end
    #       #Note: It is necessary to exclude the current unique arrangement from the 
    #       #      list, since it can be generated by periodic translations in special
    #       #      cases.
    #       #This is really slow - could do with speeding it up
    #        indices=range(iunique+1,len(all_perms))
    #        indices.reverse()
    #        for iperm in indices:
    #            if (numpy.roll(unique_perm,itranslate,axis=0)==all_perms[iperm]).all():
    #                all_perms=numpy.delete(all_perms,iperm,axis=0)
    #    length=len(all_perms)
    #    iunique=iunique+1
    #endtime=time.time()
    #print "Time taken %10.5f s" % (endtime-starttime)
    #Also tested searching using
    #iremove=numpy.where(numpy.sum(numpy.sum(numpy.fabs(all_perms-periodic_perm),axis=-1),axis=-1)==0)
    #which was slower
    #Faster way
    starttime=time.time()
    iunique=0
    length=len(all_perms)
    while iunique<length:
        unique_perm=all_perms[iunique]
        #Find periodic translations of this arrangement
        periodic_perms=[]
        for itranslate in range(1,nblocks):
            tmp_perm=numpy.roll(unique_perm,itranslate,axis=0)
            if not (tmp_perm==unique_perm).all():
                periodic_perms.append(tmp_perm)
                del tmp_perm
        periodic_perms=numpy.array(periodic_perms)
        #Find each periodic variation and remove it
        for periodic_perm in periodic_perms:
            iremove=numpy.argwhere(numpy.all(numpy.all(all_perms==periodic_perm,axis=-1),axis=-1))
            if len(iremove)==1 and int(iremove)!=iunique:
                all_perms=numpy.delete(all_perms,int(iremove),axis=0)
        length=len(all_perms)
        iunique=iunique+1
    endtime=time.time()
    print "Time taken %10.5f s" % (endtime-starttime)
    print "Building %d unique structures" % len(all_perms)
    #Build atoms objects from the unique arrangements
    structures=[]
    for arrangement in all_perms:
        #Build structure, starting with first A-terminated block
        structure=AMods[arrangement[0,0]]
        #Then first B-terminated block
        tmp_atoms=BMods[arrangement[0,1]].copy()
        tmp_atoms.translate(structure.cell[2])
        structure=structure+tmp_atoms
        structure.cell[2]=structure.cell[2]+BMods[arrangement[0,1]].cell[2]
        #Then all other blocks in similar order
        for iblock in range(1,nblocks):
            tmp_atoms=AMods[arrangement[iblock,0]].copy()
            tmp_atoms.translate(structure.cell[2])
            structure=structure+tmp_atoms
            structure.cell[2]=structure.cell[2]+AMods[arrangement[iblock,0]].cell[2]
            tmp_atoms=BMods[arrangement[iblock,1]].copy()
            tmp_atoms.translate(structure.cell[2])
            structure=structure+tmp_atoms
            structure.cell[2]=structure.cell[2]+BMods[arrangement[iblock,1]].cell[2]
            del tmp_atoms
        #Order structure according to atomic number
        structure=order_atoms(structure)
        #Add to structures images object
        structures.append(structure)
    #Rattle atoms if necessary
    if rattle_dist>0.:
        for istruct in range(len(structures)):
            structures[istruct].rattle(stdev=rattle_dist,seed=istruct)
    if write_format!='':
        write_structures(all_perms,structures,write_format,labelling,shells)
    return all_perms,structures

def write_res(fileobj, images, shells=None, name="Structure"):
    """Function to write out structure in a file readable by GULP"""
    #Test if fileobj is a file object or a path
    if isinstance(fileobj, str):
        fileobj = open(fileobj, 'w')
    #If images or shells are not already lists, make them one
    if not isinstance(images, (list, tuple)):
        images = [images,]
    if not isinstance(shells, (list, tuple)):
    	shells = [shells,]
    #Write out information for each atoms object in images
    for atoms in images:
    	### writing structure headers ###
    	symbols = atoms.get_chemical_symbols()
        fileobj.write('%s\n\n' % name )                               
        fileobj.write('# number of atoms = %d\n\n' % len(atoms)) 
        ### writing unitcell vectors ###
        fileobj.write('vectors\n')                              
        for vec in atoms.get_cell():                            
            fileobj.write('  %11.6f %11.6f %11.6f\n'% tuple(vec))
        ### writing core atomic co-ordinates ###
        fileobj.write('fractional\n')
        for atoms in images:
            for s,(x, y, z) in zip(symbols, atoms.get_scaled_positions()):
                fileobj.write('%-2s core %22.6f %22.6f %22.6f\n' % (s, x, y, z))
        ### writing shell atoms ###
        if shells !=None:
            for atoms in images:
                for s,(x, y, z) in zip(symbols, atoms.get_scaled_positions()):
                    if shells.count(s)>0:
	                fileobj.write('%-2s shel %22.6f %22.6f %22.6f\n' % (s, x, y, z))
